#!/usr/bin/perl -w

#
# Port status monitoring daemon
#

# A special note: out of nowhere on a server, printing to STDOUT (a closed
# filehandle after forking) after using LWP::Simple::get caused a segfault.
# Turns out the handle used for the socket re-used the fileno. for STDOUT,
# and when its handle was closed then STDOUT pointed nowhere.
#
# So every print or warn after the fork goes into the function "debug" now;
# if running in daemon mode, the prints are ignored; otherwise they're sent
# to STDOUT.

use strict;
use warnings;
use Mail::Sender;
use LWP::Simple;
use Sys::Hostname;
use subs "debug";

END {
	# why??
	open (WHY, ">>/tmp/port_status.crashed");
	print WHY localtime(time()) . "\n"
		. "Script exited.\n"
		. "\$\@ = $@; \$! = $!; \$\? = $?; \@\_ = @_\n\n";
	close (WHY);
}

sub debug {
	my $line = shift;
	return unless (@ARGV);
	print $line;
}

my $smtp         = 'localhost'; # Mail server
my $server_id    = undef;       # We'll figure this one out later.
my $interval     = 60*60;       # Check once per hour.
my $udpTolerance = 3;           # How many times we can have UDP flakiness before we e-mail
my $udpFailures  = 0;           # Don't change this here.
my $product      = 'PBXtra';
my $domain       = 'pbxtra.fonality.com';
my $cp           = "cp.fonality.com";

# Debug mode.
my $customCP = 0;
if (@ARGV)
{
	print "Entering debug mode. The process won't fork. Possible debug mode parameters:\n"
		. "$0 server=1358 cp=cp.test8.fonality.com\n";

	foreach my $a (@ARGV)
	{
		if ($a =~ /server=(\d+)$/i)
		{
			$server_id = $1;
		}
		elsif ($a =~ /cp=(.+?)$/i)
		{
			$customCP = 1;
			$cp = $1;
		}
	}
}

# Figure out our server ID.
unless (defined $server_id)
{
	my $globals = "/etc/asterisk/fonality/globals.conf";
	unless (open(GLOBALS, $globals))
	{
		die "Can't figure out my server ID: $!";
	}
	while (<GLOBALS>)
	{
		if (/^IAXINFO\d?=s(\d{4,6}):([^;\s]+)/)
		{
			$server_id = $1;
			last;
		}
	}
}

# Find what product this is.
if ($server_id < 100000)
{
	$product = 'PBXtra';
	$domain  = 'pbxtra.fonality.com';
	$cp      = 'cp.fonality.com';
}
else
{
	$product = 'trixbox';
	$domain  = 'trixbox.fonality.com';
	$cp      = 'cp.trixbox.com';
}

# Daemonize.
unless (@ARGV)
{
	my $pid = fork();
	if ($pid)
	{
		print "Backgrouding as PID $pid.\n";
		exit(0);
	}
	close (STDOUT);
	close (STDIN);
	close (STDERR);
}

# Enter the main loop.
my $lastCheck = 0;
main();
sub main
{
	while (1)
	{
		# Is it time to do another port check?
		if (time() > $lastCheck)
		{
			# Read our alerts2.conf.
			debug "Reading alerts2.conf\n";
			my %conf = read_alerts();

			# Do we even WANT to monitor ports?
			my $doCheck = 0;
			if (scalar keys %conf)
			{
				if ($conf{rhud_enable} == 1 || $conf{remote_enable} == 1)
				{
					# Make sure we have somebody to e-mail.
					if (length $conf{rhud_email} || length $conf{remote_email})
					{
						# Good, let's do the check.
						$doCheck = 1;
					}
				}
			}

			$lastCheck = time() + $interval;

			# Checking?
			if ($doCheck)
			{
				# Get our port scanning page.
				debug "Beginning port test...\n";
				my $url = "http://$cp/daemon.cgi?service=testports&server=$server_id";
				debug "Grab $url\n";
				my $scan = get $url;
				debug "Result: $scan\n";

				# Process the results.
				my $alertRemote = 0;
				my $alertHud = 0;
				my $results = {};
				if (length $scan)
				{
					my @lines = split(/\n/, $scan);
					foreach my $line (@lines)
					{
						if ($line =~ /^UDP/)
						{
							# UDP port results.
							my ($port,$percent,$received,$sent) = $line =~ /^UDP (\d+): (\d+)% \((\d+)\/(\d+)\)/;
							$results->{udp}->{$port} = {
								port     => $port,
								percent  => $percent,
								sent     => $sent,
								received => $received,
							};

							debug "UDP Port $port: $percent% (sent: $sent; recv: $received)\n";

							# Was this a failure?
							if ($percent < 100)
							{
								# Some packet loss. This can be tolerated a little bit.
								$udpFailures++;
								if ($udpFailures > $udpTolerance)
								{
									$alertRemote = 1;
									$udpFailures = 0;
								}
							}
						}
						elsif ($line =~ /^TCP/)
						{
							# TCP port results.
							my ($port,$result) = $line =~ /^TCP (\d+): (\w+)$/;
							$results->{tcp}->{$port} = {
								port   => $port,
								status => $result,
							};

							# Skip the HUD 2 port number.
							if ($port == 6600) {
								debug "Skipping HUD2 port results (port $port)\n";
								next;
							}

							debug "TCP port $port: $result\n";

							# Was this a failure?
							if ($result =~ /ERR/)
							{
								# The port was unreachable.
								$alertHud = 1;
							}
						}
					}

					# Do we need to send e-mail?
					if ($alertRemote || $alertHud)
					{
						my $email_product = $conf{brandingname} ? $conf{brandingname} : $product;

						# What do we need to alert about?
						if ($alertRemote && $conf{remote_enable} && length $conf{remote_email})
						{
							send_email (
								subject => "$email_product Remote Phones Alert: Server $server_id",
								message => "Your remote phone ports are having trouble. See the details in "
									. "Admin Panel->Status->diagnostics.",
								to      => $conf{remote_email},
							);
						}
						if ($alertHud && $conf{rhud_enable} && length $conf{rhud_email})
						{
							send_email (
								subject => "$email_product Remote HUD Alert: Server $server_id",
								message => "Your remote HUD port is unreachable. See more details in the "
									. "Admin Panel->Status->diagnostics.",
								to      => $conf{rhud_email},
							);
						}
					}
				}
				else
				{
					# No results?
					debug "Couldn't get a result from $cp!";
				}
			}
		
		}

		sleep 30;
	}
}

sub read_alerts
{
	# Read the alerts2.conf file.
	my $alerts = "/etc/asterisk/alerts2.conf";
	$alerts = "/etc/fonality/$server_id/alerts2.conf" unless -f $alerts;
	unless (open(ALERTS, $alerts))
	{
		debug "Can't read alerts2.conf: $!";
		return;
	}

	my %conf = ();
	while (<ALERTS>)
	{
		chomp;
		s/^\s+//g;
		s/\s+$//g;
		next unless $_ =~ /=/;
		my ($var,$value) = split(/=/, $_, 2);
		$var =~ s/\s+$//g;
		$value =~ s/^\s+//g;
		$conf{$var} = $value;
	}

	close (ALERTS);

	# Return.
	return %conf;
}

sub send_email
{
	my %mail = @_;

	# Log this first.
	addlog($mail{message});

	# Verify we got a valid e-mail.
	return unless $mail{to} =~ /\@/;

	# Send it.
	eval
	{
		my $sender = new Mail::Sender { smtp => 'localhost', from => "$product\@s$server_id.$domain" };
		$sender->MailMsg ({
			to => $mail{to},
			subject => $mail{subject},
			msg     => $mail{message},
		});
	};
	if ($@)
	{
		addlog("Couldn't deliver e-mail: $@");
	}
}

sub addlog
{
	my $msg = shift;
	unless (open (LOG, ">>/var/log/alerts.log"))
	{
		return;
	}
	print LOG localtime() . "\n";
	print LOG $msg . "\n\n";
	close (LOG);
}
